model: remove ModelChecker::is_infeasible_ignoreRMW()
[c11tester.git] / model.h
diff --git a/model.h b/model.h
index 04e191bed53340819ee1fa87c326bf9bcfbc89a6..c76a28972bf4e6f9fac715386a0a69cb1f4956e3 100644 (file)
--- a/model.h
+++ b/model.h
@@ -5,7 +5,6 @@
 #ifndef __MODEL_H__
 #define __MODEL_H__
 
-#include <list>
 #include <vector>
 #include <cstddef>
 #include <ucontext.h>
@@ -68,7 +67,8 @@ struct execution_stats {
 };
 
 struct PendingFutureValue {
-       ModelAction *writer;
+       PendingFutureValue(ModelAction *writer, ModelAction *act) : writer(writer), act(act) { }
+       const ModelAction *writer;
        ModelAction *act;
 };
 
@@ -76,7 +76,11 @@ struct PendingFutureValue {
 struct release_seq {
        /** @brief The acquire operation */
        ModelAction *acquire;
-       /** @brief The head of the RMW chain from which 'acquire' reads; may be
+       /** @brief The read operation that may read from a release sequence;
+        *  may be the same as acquire, or else an earlier action in the same
+        *  thread (i.e., when 'acquire' is a fence-acquire) */
+       const ModelAction *read;
+       /** @brief The head of the RMW chain from which 'read' reads; may be
         *  equal to 'release' */
        const ModelAction *rf;
        /** @brief The head of the potential longest release sequence chain */
@@ -91,101 +95,102 @@ public:
        ModelChecker(struct model_params params);
        ~ModelChecker();
 
+       void run();
+
        /** @returns the context for the main model-checking system thread */
        ucontext_t * get_system_context() { return &system_context; }
 
        void print_summary() const;
 #if SUPPORT_MOD_ORDER_DUMP
-       void dumpGraph(char *filename);
+       void dumpGraph(char *filename) const;
 #endif
 
        void add_thread(Thread *t);
        void remove_thread(Thread *t);
        Thread * get_thread(thread_id_t tid) const;
-       Thread * get_thread(ModelAction *act) const;
+       Thread * get_thread(const ModelAction *act) const;
 
        bool is_enabled(Thread *t) const;
        bool is_enabled(thread_id_t tid) const;
 
        thread_id_t get_next_id();
        unsigned int get_num_threads() const;
-       Thread * get_current_thread();
+       Thread * get_current_thread() const;
 
-       int switch_to_master(ModelAction *act);
-       ClockVector * get_cv(thread_id_t tid);
-       ModelAction * get_parent_action(thread_id_t tid);
-       bool next_execution();
-       bool isfeasible() const;
-       bool isfeasibleotherthanRMW() const;
-       bool isfinalfeasible() const;
+       uint64_t switch_to_master(ModelAction *act);
+       ClockVector * get_cv(thread_id_t tid) const;
+       ModelAction * get_parent_action(thread_id_t tid) const;
        void check_promises_thread_disabled();
-       void mo_check_promises(thread_id_t tid, const ModelAction *write);
-       void check_promises(thread_id_t tid, ClockVector *old_cv, ClockVector * merge_cv);
-       void get_release_seq_heads(ModelAction *act, rel_heads_list_t *release_heads);
-       void finish_execution();
+       void mo_check_promises(thread_id_t tid, const ModelAction *write, const ModelAction * read);
+       void check_promises(thread_id_t tid, ClockVector *old_cv, ClockVector *merge_cv);
        bool isfeasibleprefix() const;
 
        bool assert_bug(const char *msg);
        void assert_user_bug(const char *msg);
 
-       bool is_deadlocked() const;
-       bool is_complete_execution() const;
-       void print_stats() const;
-
-       void set_bad_synchronization();
-
        const model_params params;
-       Node * get_curr_node();
+       Node * get_curr_node() const;
 
        MEMALLOC
 private:
        /** The scheduler to use: tracks the running/ready Threads */
        Scheduler *scheduler;
 
-       bool sleep_can_read_from(ModelAction * curr, const ModelAction *write);
-       bool thin_air_constraint_may_allow(const ModelAction * writer, const ModelAction *reader);
-       bool mo_may_allow(const ModelAction * writer, const ModelAction *reader);
+       bool sleep_can_read_from(ModelAction *curr, const ModelAction *write);
+       bool thin_air_constraint_may_allow(const ModelAction *writer, const ModelAction *reader);
+       bool mo_may_allow(const ModelAction *writer, const ModelAction *reader);
        bool has_asserted() const;
        void set_assert();
+       void set_bad_synchronization();
        bool promises_expired() const;
        void execute_sleep_set();
-       void wake_up_sleeping_actions(ModelAction * curr);
+       void wake_up_sleeping_actions(ModelAction *curr);
        modelclock_t get_next_seq_num();
 
+       bool next_execution();
        void set_current_action(ModelAction *act);
-       Thread * check_current_action(ModelAction *curr);
+       ModelAction * check_current_action(ModelAction *curr);
        bool initialize_curr_action(ModelAction **curr);
        bool process_read(ModelAction *curr, bool second_part_of_rmw);
        bool process_write(ModelAction *curr);
+       bool process_fence(ModelAction *curr);
        bool process_mutex(ModelAction *curr);
        bool process_thread_action(ModelAction *curr);
        void process_relseq_fixup(ModelAction *curr, work_queue_t *work_queue);
+       bool read_from(ModelAction *act, const ModelAction *rf);
        bool check_action_enabled(ModelAction *curr);
 
-       bool take_step();
+       bool take_step(ModelAction *curr);
 
        void check_recency(ModelAction *curr, const ModelAction *rf);
        ModelAction * get_last_conflict(ModelAction *act);
        void set_backtracking(ModelAction *act);
        Thread * get_next_thread(ModelAction *curr);
+       bool set_latest_backtrack(ModelAction *act);
        ModelAction * get_next_backtrack();
        void reset_to_initial_state();
        bool resolve_promises(ModelAction *curr);
        void compute_promises(ModelAction *curr);
        void compute_relseq_breakwrites(ModelAction *curr);
 
-       void check_curr_backtracking(ModelAction * curr);
+       void check_curr_backtracking(ModelAction *curr);
        void add_action_to_lists(ModelAction *act);
        ModelAction * get_last_action(thread_id_t tid) const;
-       ModelAction * get_last_seq_cst(ModelAction *curr) const;
+       ModelAction * get_last_fence_release(thread_id_t tid) const;
+       ModelAction * get_last_seq_cst_write(ModelAction *curr) const;
+       ModelAction * get_last_seq_cst_fence(thread_id_t tid, const ModelAction *before_fence) const;
        ModelAction * get_last_unlock(ModelAction *curr) const;
        void build_reads_from_past(ModelAction *curr);
        ModelAction * process_rmw(ModelAction *curr);
        void post_r_modification_order(ModelAction *curr, const ModelAction *rf);
        bool r_modification_order(ModelAction *curr, const ModelAction *rf);
        bool w_modification_order(ModelAction *curr);
+       void get_release_seq_heads(ModelAction *acquire, ModelAction *read, rel_heads_list_t *release_heads);
        bool release_seq_heads(const ModelAction *rf, rel_heads_list_t *release_heads, struct release_seq *pending) const;
        bool resolve_release_sequences(void *location, work_queue_t *work_queue);
+       void add_future_value(const ModelAction *writer, ModelAction *reader);
+
+       ModelAction * new_uninitialized_action(void *location) const;
 
        ModelAction *diverge;
        ModelAction *earliest_diverge;
@@ -219,6 +224,7 @@ private:
        std::vector< struct release_seq *, SnapshotAlloc<struct release_seq *> > *pending_rel_seqs;
 
        std::vector< ModelAction *, SnapshotAlloc<ModelAction *> > *thrd_last_action;
+       std::vector< ModelAction *, SnapshotAlloc<ModelAction *> > *thrd_last_fence_release;
        NodeStack *node_stack;
 
        /** Private data members that should be snapshotted. They are grouped
@@ -249,9 +255,17 @@ private:
        struct execution_stats stats;
        void record_stats();
 
+       void print_infeasibility(const char *prefix) const;
+       bool is_feasible_prefix_ignore_relseq() const;
+       bool is_infeasible() const;
+       bool is_deadlocked() const;
+       bool is_complete_execution() const;
        bool have_bug_reports() const;
        void print_bugs() const;
        void print_execution(bool printbugs) const;
+       void print_stats() const;
+
+       friend void user_main_wrapper();
 };
 
 extern ModelChecker *model;